/* ====================================================================
 * The Apache Software License, Version 1.1
 *
 * Copyright (c) 2000 The Apache Software Foundation.  All rights
 * reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 *
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in
 *    the documentation and/or other materials provided with the
 *    distribution.
 *
 * 3. The end-user documentation included with the redistribution,
 *    if any, must include the following acknowledgment:
 *       "This product includes software developed by the
 *        Apache Software Foundation (http://www.apache.org/)."
 *    Alternately, this acknowledgment may appear in the software itself,
 *    if and wherever such third-party acknowledgments normally appear.
 *
 * 4. The names "Apache" and "Apache Software Foundation" must
 *    not be used to endorse or promote products derived from this
 *    software without prior written permission. For written
 *    permission, please contact apache@apache.org.
 *
 * 5. Products derived from this software may not be called "Apache",
 *    nor may "Apache" appear in their name, without prior written
 *    permission of the Apache Software Foundation.
 *
 * THIS SOFTWARE IS PROVIDED ``AS IS'' AND ANY EXPRESSED OR IMPLIED
 * WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
 * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
 * DISCLAIMED.  IN NO EVENT SHALL THE APACHE SOFTWARE FOUNDATION OR
 * ITS CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF
 * USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
 * OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT
 * OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 * ====================================================================
 *
 * This software consists of voluntary contributions made by many
 * individuals on behalf of the Apache Software Foundation.  For more
 * information on the Apache Software Foundation, please see
 * <http://www.apache.org/>.
 *
 * Portions of this software are based upon public domain software
 * originally written at the National Center for Supercomputing Applications,
 * University of Illinois, Urbana-Champaign.
 */

#include "mod_cache.h"

#define CORE_PRIVATE

#include "http_log.h"

apr_status_t file_cache_close(ap_cache_handle_t *h)
{
    apr_destroy_pool(h->pool);
    return APR_SUCCESS;
}
apr_status_t file_cache_el_hdr_walk(ap_cache_el *el, 
                                   int (*comp)(void *, const char *, const char *), void *rec, va_list args)
{
    cache_req *cq = (cache_req *)el;
    if(!cq->hdrs && !file_cache_get_hdrs(cq))
        return errno;
    apr_table_vdo(comp, rec, cq->hdrs, args);
    return APR_SUCCESS;
}
apr_status_t file_cache_el_hdr(ap_cache_el *el, const char *name, const char *val, ap_cache_query flag)
{
    cache_req *cq = (cache_req *)el;

    if(!cq->hdrs && !file_cache_get_hdrs(cq))
        return errno;
    switch(flag)
    {
    case AP_CACHE_CREATE:
        apr_table_add(cq->hdrs, name, val);
        break;
    case AP_CACHE_CHANGE:
        apr_table_set(cq->hdrs, name, val);
        break;
    case AP_CACHE_REMOVE:
        apr_table_unset(cq->hdrs, name);
        break;
    default:
        return APR_BADARG;
    }
    return APR_SUCCESS;
}
apr_status_t file_cache_el_data(ap_cache_el *el, apr_file_t **b)
{
    cache_req *cq = (cache_req *)el;
    *b = NULL;
    if(!cq->fp && !file_cache_get_data(cq))
        return errno;
    *b = cq->fp;
    return APR_SUCCESS;
}
apr_status_t file_cache_reset(ap_cache_el *el, ap_cache_part f)
{
    cache_req *cq = (cache_req *)el;
    if(f == AP_CACHE_DATA && cq->fp)
    {
        cq->state &= ~FC_READ_MY_DATA; /* re-read my data later */
        apr_close(cq->fp);
        cq->fp = NULL;
    }
    else if(f == AP_CACHE_HEADER && cq->hdrs)
    {
        cq->state &= ~FC_READ_HEADERS; /* re-read headers later */
        cq->hdrs = NULL; 
    }
    return APR_SUCCESS;
}
        
/* module hook and callback setup */
static apr_status_t cache_init(ap_cache_handle_t **h, const char *desc, server_rec *server)
{
    apr_pool_t *my_pool;
    ap_cache_handle_t *han;

    if(!h) return APR_BADARG;

    apr_create_pool(&my_pool, server->process->pool);
    han = (*h) = apr_pcalloc(my_pool, sizeof(ap_cache_handle_t));
    
    memset(han, '\0', sizeof(ap_cache_handle_t));
    han->pool = my_pool;
    han->server = server;
    han->meth.cache_garbage_coll = file_cache_gc; /* file_garbage.c */
    han->meth.cache_element = file_cache_element; /* file_cache.c */
    han->meth.cache_el_final = file_cache_el_final; /* file_cache.c */
    han->meth.cache_close = file_cache_close; /* here */
    han->meth.cache_el_header_walk = file_cache_el_hdr_walk; /* here */
    han->meth.cache_el_hdr = file_cache_el_hdr; /* here */
    han->meth.cache_el_data = file_cache_el_data; /* here */
    han->meth.cache_el_reset = file_cache_reset; /* here */

    return APR_SUCCESS;
}


/* -------------------------------------------------------------- */
/* Setup configurable data */

static void *create_cache_config(apr_pool_t *p, server_rec *s)
{
    cache_server_conf *ps = apr_pcalloc(p, sizeof(cache_server_conf));
    ps->root = NULL;
    ps->space = DEFAULT_CACHE_SPACE;
    ps->maxexpire = DEFAULT_CACHE_MAXEXPIRE;
    ps->defaultexpire = DEFAULT_CACHE_EXPIRE;
    ps->lmfactor = DEFAULT_CACHE_LMFACTOR;
    ps->gcinterval = -1;
    /* at these levels, the cache can have 2^18 directories (256,000)  */
    ps->dirlevels = 3;
    ps->dirlength = 1;
    return ps;
}


static const char
*set_cache_size(cmd_parms *parms, char *struct_ptr, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    int val;

    if (sscanf(arg, "%d", &val) != 1)
    return "CacheSize value must be an integer (kBytes)";
    psf->space = val;
    return NULL;
}

static const char
*set_cache_root(cmd_parms *parms, void *dummy, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);

    psf->root = ap_server_root_relative(parms->pool, arg);

    return NULL;
}

static const char
*set_cache_factor(cmd_parms *parms, void *dummy, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    double val;

    if (sscanf(arg, "%lg", &val) != 1)
        return "CacheLastModifiedFactor value must be a float";
    psf->lmfactor = val;

    return NULL;
}

static const char
*set_cache_maxex(cmd_parms *parms, void *dummy, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    double val;

    if (sscanf(arg, "%lg", &val) != 1)
        return "CacheMaxExpire value must be a float";
    psf->maxexpire = (apr_time_t) (val * MSEC_ONE_HR);
    return NULL;
}

static const char
*set_cache_defex(cmd_parms *parms, void *dummy, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    double val;

    if (sscanf(arg, "%lg", &val) != 1)
        return "CacheDefaultExpire value must be a float";
    psf->defaultexpire = (apr_time_t) (val * MSEC_ONE_HR);
    return NULL;
}

static const char
*set_cache_gcint(cmd_parms *parms, void *dummy, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    double val;

    if (sscanf(arg, "%lg", &val) != 1)
        return "CacheGcInterval value must be a float";
    psf->gcinterval = (apr_time_t) (val * (double) MSEC_ONE_HR);
    return NULL;
}

static const char
*set_cache_dirlevels(cmd_parms *parms, char *struct_ptr, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    int val;

    val = atoi(arg);
    if (val < 1)
        return "CacheDirLevels value must be an integer greater than 0";
    if (val * psf->dirlength > CACHEFILE_LEN)
    return "CacheDirLevels*CacheDirLength value must not be higher than 20";
    psf->dirlevels = val;
    return NULL;
}

static const char
*set_cache_dirlength(cmd_parms *parms, char *struct_ptr, char *arg)
{
    cache_server_conf *psf = ap_get_module_config(parms->server->module_config, &cache_module);
    int val;

    val = atoi(arg);
    if (val < 1)
        return "CacheDirLength value must be an integer greater than 0";
    if (val * psf->dirlevels > CACHEFILE_LEN)
        return "CacheDirLevels*CacheDirLength value must not be higher than 20";
    psf->dirlength = val;
    return NULL;
}

static const command_rec cache_cmds[] =
{
    {"CacheRoot", set_cache_root, NULL, RSRC_CONF, TAKE1,
     "The directory to store cache files"},
    {"CacheSize", set_cache_size, NULL, RSRC_CONF, TAKE1,
     "The maximum disk space used by the cache in Kb"},
    {"CacheMaxExpire", set_cache_maxex, NULL, RSRC_CONF, TAKE1,
     "The maximum time in hours to cache a document"},
    {"CacheDefaultExpire", set_cache_defex, NULL, RSRC_CONF, TAKE1,
     "The default time in hours to cache a document"},
    {"CacheLastModifiedFactor", set_cache_factor, NULL, RSRC_CONF, TAKE1,
     "The factor used to estimate Expires date from LastModified date"},
    {"CacheGcInterval", set_cache_gcint, NULL, RSRC_CONF, TAKE1,
     "The interval between garbage collections, in hours"},
    {"CacheDirLevels", set_cache_dirlevels, NULL, RSRC_CONF, TAKE1,
     "The number of levels of subdirectories in the cache"},
    {"CacheDirLength", set_cache_dirlength, NULL, RSRC_CONF, TAKE1,
     "The number of characters in subdirectory names"},
    {NULL}
};

static void
register_hooks(void)
{
    /* cache initializer */
    ap_hook_cache_init(cache_init, NULL, NULL, AP_HOOK_FIRST);
}

module AP_MODULE_DECLARE_DATA cache_module =
{
    STANDARD20_MODULE_STUFF,
    NULL,            /* create per-directory config structure */
    NULL,            /* merge per-directory config structures */
    create_cache_config,    /* create per-server config structure */
    NULL,            /* merge per-server config structures */
    cache_cmds,            /* command apr_table_t */
    NULL,            /* handlers */
    register_hooks
};
